EC2_P_256_NUMBER_LENGTH = 32
EC2_P_384_NUMBER_LENGTH = 48
EC2_P_521_NUMBER_LENGTH = 66
OKP_ED25519_NUMBER_LENGTH = 32
OKP_ED448_NUMBER_LENGTH = 57

# https://android.googlesource.com/platform/hardware/libhardware/+/master/include/hardware/keymaster_defs.h
KM_ORIGIN_GENERATED = 0
KM_PURPOSE_SIGN = 2
